# -*- python -*-
# $RCSfile: oofcanvas3d.py,v $
# $Revision: 1.4 $
# $Author: vrc $
# $Date: 2007-09-27 21:23:11 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

import gtk
import gtk.gtkgl
from gtk import gdk
import vtk
import math
from ooflib.SWIG.common import ooferror
from ooflib.common import color
from ooflib.common import primitives
from ooflib.common import mainthread
from ooflib.common.IO import voxelpicker
from ooflib.SWIG.common.IO import bitoverlay
from ooflib.SWIG.image import oofimage3d
from ooflib.SWIG.image import resampleimage
#from ooflib.image import oofimage3d

# TODO: NEEDS MAJOR CLEANUP - will be done when we convert to c++

class OOFCanvas3DLayer:

    def __init__(self, canvas):
        self.canvas = canvas
        self.resample = None
        self.inputPort = None
        self.image = None
        self.showing = True

    def make_current(self):
        self.canvas.current_layer = self

    def clear(self):
        self.detach_from_pipeline()
        self.resample=None
        self.image=None

    def show(self):
        if not self.showing: # don't do anything if layer already showing
            self.showing = True
            self.attach_to_pipeline()
           
    def hide(self):
        if self.showing: # don't do anything if layer already hidden
            self.showing = False
            self.detach_from_pipeline()

    def raise_layer(self, howfar):
        index = self.get_index()
        self.canvas.layers.remove(self)
        self.canvas.layers.insert(index-howfar, self)

    def lower_layer(self, howfar):
        index = self.get_index()
        self.canvas.layers.remove(self)
        self.canvas.layers.insert(index+howfar, self)

    def add_image(self, image):
        self.image = image
        self.attach_to_pipeline()

    def get_index(self):
        return self.canvas.layers.index(self)

    def add_resample(self, resample):
        self.resample = resample
        # get the next lowest image layer and make that our input if
        # we don't already have one
        self.find_inputPort()
        if self.inputPort is not None:
            self.resample.SetInputConnection(self.inputPort)
            self.image = self.resample.GetOutput()
        else:
            # This is reached if we lower a resampled layer to where
            # there is no lower layer.  In that case, do nothing.
            pass

        self.attach_to_pipeline()

    def find_inputPort(self):
        index = self.get_index()
        self.inputPort = None
        for layer in self.canvas.layers[index+1:]:
            if layer.image is not None and self.inputPort is None:
                self.inputPort = layer.image.GetProducerPort()
        

    def attach_to_pipeline(self):
        # find next layer with resample member and inject our input
        if self.image is not None:
            index = self.get_index()
            under_opaque_layer = False
            for layer in self.canvas.layers[:index]:
                if layer.resample is not None and not under_opaque_layer:
                    layer.inputPort = self.image.GetProducerPort()
                    layer.resample.SetInputConnection(layer.inputPort)
                elif layer.resample is None and layer.image is not None:
                    under_opaque_layer = True


    def detach_from_pipeline(self):
        self.find_inputPort()
        if self.image is not None:
            index = self.get_index()
            for layer in self.canvas.layers[:index]:
                if layer.resample is not None:
                    layer.inputPort = self.inputPort
                    layer.resample.SetInputConnection(layer.inputPort)
        if self.resample is not None:
            self.resample.RemoveAllInputs()
        

    def update_image(self):
        if self.resample is not None:
            if self.inputPort is None:
                self.image = None
                self.detach_from_pipeline()
            else:
                self.image = self.resample.GetOutput()
                self.attach_to_pipeline()


    def destroy(self):
        self.clear()


        
        

class OOFCanvas3D(gtk.gtkgl.DrawingArea):    

    def __init__(self, *args):
        gtk.gtkgl.DrawingArea.__init__(self)

        self._RenderWindow = vtk.vtkRenderWindow()
        self.__Created = 0
        self.__Exposed = 0

        # used by the actors
        self._DesiredUpdateRate = 15
        self._StillUpdateRate = 0.0001

        self.connectSignals()

        # need this to be able to handle key_press events.
        self.set_flags(gtk.CAN_FOCUS)
        # default size
        self.set_size_request(300, 300)

        # For now, we are only using one renderer and one camera
        self._Renderer =  vtk.vtkRenderer()
        self._RenderWindow.AddRenderer(self._Renderer)
        self._Camera = self._Renderer.GetActiveCamera()


        self._ViewportCenterX = 0
        self._ViewportCenterY = 0

        self._ClippingRange = (0,0)
        
        self._Picker = voxelpicker.VoxelPicker()
        self._Picker.SetTolerance(.5)
        
        self._OldFocus = None

        # these record the previous mouse position
        self._LastX = 0
        self._LastY = 0

        self.empty = True
        self.current_layer = None
        #self.top_volume_layer = None
        self.layers = []

        self.mouse_callback = None

        self.fillColor = None

        # objects used by volume actor TODO: For now, we are only
        # assigning color transfer functions for the first three
        # components.  Something strange is going on with vtk as far
        # as rendering the fourth component as opacity -- the same
        # code that worked before we converted the oofimage3d class to
        # C++ doesn't work now.  The opacity options will still be in
        # the interface but do nothing.  This should be debugged when
        # we convert this class to C++ and get rid of gtkglext.  I
        # think it might have something to do with gtkglext and/or
        # threading issues.
        self.volproperty = vtk.vtkVolumeProperty()
        #self.volproperty.IndependentComponentsOff()
        self.red = vtk.vtkColorTransferFunction()
        self.red.AddRGBSegment(0,0,0,0,255,1,0,0)
        self.volproperty.SetColor(0,self.red)
        self.green = vtk.vtkColorTransferFunction()
        self.green.AddRGBSegment(0,0,0,0,255,0,1,0)
        self.volproperty.SetColor(1,self.green)
        self.blue = vtk.vtkColorTransferFunction()
        self.blue.AddRGBSegment(0,0,0,0,255,0,0,1)
        self.volproperty.SetColor(2,self.blue)
#         self.opacity = vtk.vtkPiecewiseFunction()
##         self.opacity.AddSegment(0,.5,255,.5)
##         self.volproperty.SetScalarOpacity(self.opacity)


    def set_mouse_callback(self, callback):
        self.mouse_callback = callback
        self.connect("event", self.mouse_event)

    def mouse_event(self, wid, event):
        # not sure if this is doing everything the C++ 2D version does...
        # check that we are not dragging another window around....
        eventtype = None
        if event.type == gdk.BUTTON_PRESS:
            eventtype = "down"
        if event.type == gdk.BUTTON_RELEASE:
            eventtype = "up"
        if event.type == gdk.MOTION_NOTIFY:
            eventtype = "move"

        if eventtype != None:
            shift = event.state & gdk.SHIFT_MASK
            ctrl = event.state & gdk.CONTROL_MASK
            (x, y) = self.get_pointer()
            self.mouse_callback(eventtype, x, y, shift, ctrl)
        
    def newLayer(self):
        self.current_layer = OOFCanvas3DLayer(self)
        self.layers.insert(0,self.current_layer)
        return self.current_layer
                    

    def is_empty(self):
        # self.empty is set to true by self.clear
        # set to false by self.draw_dot, self.draw_triangle,
        # self.draw_segment(s), self.draw_curve... and draw_image
        return self.empty

    def set_lineColor(self, color):
        pass

    def set_fillColor(self, color, alpha=1):
        self.fillColor = color


    # This should be the only function that adds a volume to the renderer.
    def update_highest_showing_volume(self):
        self.remove_all_volumes()
                
        for layer in self.layers:
            if layer.image is not None and layer.showing:

                layer.image.Update()
                mapper = vtk.vtkFixedPointVolumeRayCastMapper()
                mapper.IntermixIntersectingGeometryOn()
                mapper.SetInput(layer.image)
                volume = vtk.vtkVolume()
                volume.SetMapper(mapper)
                volume.SetProperty(self.volproperty)
        
                self._Renderer.AddVolume(volume)
                self.setSampleDistances()
                self.empty = False
                return True
                
        return False
       

    def remove_all_volumes(self):
        volumes = self._Renderer.GetVolumes()
        volumes.InitTraversal()
        for i in xrange(volumes.GetNumberOfItems()):
            volume = volumes.GetNextItem()
            self._Renderer.RemoveVolume(volume)


    def draw_voxel(self, voxel):
        #  TODO: In future versions, we will probably want to change
        # this to draw a frame around a voxel -- for voxel info.
        # Hardcoding the alphas for now

        # This is a hack, at the moment
        volumes = self._Renderer.GetVolumes()
        volumes.InitTraversal()
        for i in xrange(volumes.GetNumberOfItems()):
            volume = volumes.GetNextItem()
        bounds = volume.GetBounds()
        size = primitives.iPoint(bounds[1],bounds[3],bounds[5])

        bitmap = bitoverlay.BitmapOverlay(size, size)
        bitmap.set(voxel)
        bitmap.setColor(self.fillColor)
        bitmap.setTintAlpha(0.3)

        resample = resampleimage.ResampleImage()
        resample.SetBitmap(bitmap)
        self.current_layer.add_resample(resample)


    def draw_alpha_image(self, bitmap, location, size):
        resample = resampleimage.ResampleImage()
        resample.SetBitmap(bitmap)
        self.current_layer.add_resample(resample)
                  

    def draw_image(self, oofimage, offset, size):
        volume = oofimage.getVolumeActor()
        image = oofimage.getImageData()

        self.current_layer.add_image(image)
        self._Picker.SetImage(image)
        self.update_highest_showing_volume()
        
       
    def connectSignals(self):
        self.connect("realize", self.onRealize)
        self.connect("expose_event", self.onExpose)
        self.connect("configure_event", self.onConfigure)
        self.connect("delete_event", self.onDestroy)
        self.add_events(gdk.EXPOSURE_MASK|
                        gdk.BUTTON_PRESS_MASK |
                        gdk.BUTTON_RELEASE_MASK |
                        gdk.KEY_PRESS_MASK |
                        gdk.POINTER_MOTION_MASK |
                        gdk.POINTER_MOTION_HINT_MASK |
                        gdk.ENTER_NOTIFY_MASK |
                        gdk.LEAVE_NOTIFY_MASK)

    def getRenderWindow(self):
        return self._RenderWindow

    def getRenderer(self):
        return self._Renderer

    def getCamera(self):
        return self._Camera

##     def setDesiredUpdateRate(self, rate):
##         self._DesiredUpdateRate = rate

##     def getDesiredUpdateRate(self):
##         return self._DesiredUpdateRate 
   
##     def SetStillUpdateRate(self, rate):
##         self._StillUpdateRate = rate

##     def GetStillUpdateRate(self):
##         return self._StillUpdateRate

    def getPicker(self):
        return self._Picker

    # This should only be called by higher level classes (with the
    # exception of the expose callback).  Rendering is expensive, we
    # only want to render when we are really ready.
    def render(self):
        if self.__Exposed:
            mainthread.run(self._RenderWindow.Render)

    def update_volume_and_render_threaded(self):
        if self.__Exposed:
            self.update_highest_showing_volume()
            self._RenderWindow.Render()

    def update_volume_and_render(self):
        mainthread.run(self.update_volume_and_render_threaded)

    def onRealize(self, *args):
        if self.__Created == 0:
            # you can't get the xid without the window being realized.
            self.realize()
            win_id = str(self.widget.window.xid)
            self._RenderWindow.SetWindowInfo(win_id)
            self.__Created = 1
        return True

    def onConfigure(self, wid, event=None):
        self.widget=wid
        sz = self._RenderWindow.GetSize()
        
        if (event.width != sz[0]) or (event.height != sz[1]):
            self._RenderWindow.SetSize(event.width, event.height)
        return True

    def onExpose(self, *args):
        if not self.__Exposed:
            self.__Exposed = 1
        self.render()
        return True

    def onDestroy(self, *args):
        self.hide()
        del self._RenderWindow
        self.destroy()
        return True



    def updateXY(self,x,y):

        windowX,windowY  = self.widget.window.get_size()

        vx,vy = (0,0)
        if (windowX > 1):
            vx = float(x)/(windowX-1)
        if (windowY > 1):
            vy = (windowY-float(y)-1)/(windowY-1)
        (vpxmin,vpymin,vpxmax,vpymax) = self._Renderer.GetViewport()

        if (vx >= vpxmin and vx <= vpxmax and
            vy >= vpymin and vy <= vpymax):
            self._ViewportCenterX = float(windowX)*(vpxmax-vpxmin)/2.0\
                                    +vpxmin
            self._ViewportCenterY = float(windowY)*(vpymax-vpymin)/2.0\
                                    +vpymin

        self._LastX = x
        self._LastY = y





    def mouseTumble(self,x,y):            
        self._Camera.Azimuth(self._LastX - x)
        self._Camera.Elevation(y - self._LastY)
        self._Camera.OrthogonalizeViewUp()

        self._LastX = x
        self._LastY = y

        #self.resetClippingRange()
        #self.render()


    def mouseTrack(self,x,y):
        
        # if we click on a voxel, we will track that voxel, otherwise
        # we track the focal point.
        clickedVoxel = self.voxelInfo(x,y)
        fPoint = self._Camera.GetFocalPoint()
        if clickedVoxel is not None:
            trackPoint = clickedVoxel
        else:
            trackPoint = fPoint

        # convert track point to display coordinates
        self._Renderer.SetWorldPoint(trackPoint[0],trackPoint[1],trackPoint[2],1.0)
        self._Renderer.WorldToDisplay()
        dPoint = self._Renderer.GetDisplayPoint()

        # add mouse displacement to track point in display coordinates
        aPoint0 = (x - self._LastX)
        aPoint1 = -(y - self._LastY)
        newDPoint = (dPoint[0]+aPoint0,dPoint[1]+aPoint1,dPoint[2])

        #convert newDisplayPoint to world coordinates
        self._Renderer.SetDisplayPoint(newDPoint[0],newDPoint[1],newDPoint[2])
        self._Renderer.DisplayToWorld()
        rPoint = list(self._Renderer.GetWorldPoint())
        if (rPoint[3] != 0.0):
            rPoint[0] = rPoint[0]/rPoint[3]
            rPoint[1] = rPoint[1]/rPoint[3]
            rPoint[2] = rPoint[2]/rPoint[3]

        self._LastX = x
        self._LastY = y

        # track by displacement in display coordinates
        self.track(trackPoint[0] - rPoint[0],
                   trackPoint[1] - rPoint[1],
                   trackPoint[2] - rPoint[2])
           

    def track(self,x,y,z):
        (pPoint0,pPoint1,pPoint2) = self._Camera.GetPosition()
        (fPoint0,fPoint1,fPoint2) = self._Camera.GetFocalPoint()
        self._Camera.SetFocalPoint(x + fPoint0, 
                             y + fPoint1,
                             z + fPoint2) 
                
        self._Camera.SetPosition(x + pPoint0, 
                           y + pPoint1,
                           z + pPoint2)
        


    def mouseDolly(self,x,y):
        dollyFactor = math.pow(1.02,(0.5*(self._LastY - y)))
        self.dolly(dollyFactor)
        self._LastX = x
        self._LastY = y


    def dolly(self,dollyFactor):
        self._Camera.Dolly(dollyFactor)        
        #self.render()

    def dollyfill(self):
        # first, recenter the image
        d = self._Camera.GetDistance()
        dop = self._Camera.GetDirectionOfProjection()
        (dummy, width, dummy, height, dummy, depth) = \
                self._Renderer.ComputeVisiblePropBounds()
        fp = (float(width)/2, float(height)/2, float(depth)/2)
        self._Camera.SetFocalPoint(fp)
        self._Camera.SetPosition(fp[0]-d*dop[0],fp[1]-d*dop[1],fp[2]-d*dop[2])

        # Find the corner that is either closest to are furthest from
        # the viewport edges and dolly such that the width or height
        # of the projection is 90% of the width or height of the
        # viewport. Since this corner is likely not in the focal
        # plane, we must take into account the depth.
        largest = 0
        # one way to do this is to dolly in a loop until we
        # converge on 90%, only render after converging
        while math.fabs(largest-0.9) > 1e-3:
            largest = 0
            world_corners = self.getWorldCorners()
            for corner in world_corners:
                self._Renderer.SetWorldPoint(corner[0],corner[1],corner[2],1.0)
                self._Renderer.WorldToView()
                view = self._Renderer.GetViewPoint()
                largest = max(largest, math.fabs(view[0]), math.fabs(view[1]))

            dollyfactor =  0.9/largest

            self._Camera.Dolly(dollyfactor)        


    def reset(self):
        self._Renderer.ResetCamera()           


    def setSampleDistances(self):
        # calculate dimensions of window in voxel units
        angle = float(self._Camera.GetViewAngle())/2*(math.pi/180)
        #TODO: this gives height at focal depth, but really we want
        #the closest point in the image
        height = 2*self._Camera.GetDistance()*math.tan(angle)
        window_height = self.widget.window.get_size()[1]
        sample_distance = min(height/window_height, .2)
        # there should be only one volume, but we do it this way
        # anyway since the renderer can only return a VolumeCollection
        # type, and someday we might want to have more than one volume
        # at a time.  We only need to update the sample distances for
        # the volumes that are currently showing.
        numvolumes = self._Renderer.VisibleVolumeCount()
        volumes = self._Renderer.GetVolumes()
        volumes.InitTraversal()
        for i in xrange(numvolumes):
            volume = volumes.GetNextVolume()
            mapper = volume.GetMapper()
            mapper.SetSampleDistance(sample_distance)
            mapper.SetInteractiveSampleDistance(2*sample_distance)
            
        


    def resetClippingRange(self):
        self._Renderer.ResetCameraClippingRange()
        self._ClippingRange = self._Camera.GetClippingRange()
 
    def getClippingRange(self):
        return self._ClippingRange


    def voxelInfo(self,x,y):
        picker = self._Picker

        windowY = self.widget.window.get_size()[1]
        foundVox = picker.Pick(x,(windowY - y - 1),0.0,self._Renderer)

        if foundVox:
            pointid = picker.GetPointId()
            voxelPoint = picker.GetPickedVoxel()

            return voxelPoint

     # Clipping Plane
    ###############################################################

    def calculateclipping(self, percent):

        # figure out the distance from the camera to the nearest
        # corner and the furthest corner
        if percent != 100:
            world_corners = self.getWorldCorners()
            cp = self._Camera.GetPosition()
            camera_position = primitives.Point(cp[0],cp[1],cp[2])
            p = self._Camera.GetDirectionOfProjection()
            projection = primitives.Point(p[0],p[1],p[2])
            near = self._Camera.GetDistance()*1e20
            far = 0
            for corner in world_corners:
                # calculate distance along direction of projection to corner
                dist = projection*(corner-camera_position)
                if dist < near:
                    near = dist
                if dist > far:
                    far = dist
            newnear = far - (far-near)*(percent/100)
            self._Camera.SetClippingRange(newnear,far)
        else:
            self.resetClippingRange()
       

    def getWorldCorners(self):
        # get the corners of the visible props in world coordinates
        bounds = self._Renderer.ComputeVisiblePropBounds()
        return [primitives.Point(bounds[0],bounds[2],bounds[4]),
                primitives.Point(bounds[0],bounds[2],bounds[5]),
                primitives.Point(bounds[0],bounds[3],bounds[4]),
                primitives.Point(bounds[0],bounds[3],bounds[5]),
                primitives.Point(bounds[1],bounds[2],bounds[4]),
                primitives.Point(bounds[1],bounds[2],bounds[5]),
                primitives.Point(bounds[1],bounds[3],bounds[4]),
                primitives.Point(bounds[1],bounds[3],bounds[5])]
                


















