// -*- C++ -*-
// $RCSfile: csubproblem.h,v $
// $Revision: 1.15 $
// $Author: rlua $
// $Date: 2007-09-28 03:20:42 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#ifndef SUBPROBLEM_H
#define SUBPROBLEM_H

// Underlying C++ base class for SubProblems.  SubProblems contain all
// of the machinery for defining Fields and solving Equations on a
// FEMesh.  The FEMesh hosts the subproblems and contains the Nodes
// and Elements.  The various derived SubProblem classes differ in how
// they chose which portions of the FEMesh to solve.

class CSubProblem;
#include <vector>
#include "common/identification.h"
#include "common/lock.h"
#include "engine/femesh.h"
#include "engine/linearizedsystem.h"
#include "engine/matvec.h"

class CConjugatePair;
class CMicrostructure;
class CSRmatrix;
class CompoundField;
class Coord;
class Element;
class ElementIterator;
class Equation;
class FEMesh;
class Field;
class Flux;
class FuncNode;
class FuncNodeIterator;
class Lock;
class Material;
class NodalEquation;
class Node;
class NodeIterator;
//AMR subproblem
class NodalFluxes;
class NodalSCPatches;

// CSubProblem is derived from IdentifiedObject so that it can provide
// a simple __hash__ function, defined in csubproblem.swg.

class CSubProblem : public IdentifiedObject {
private:
  RWLock *rwlock;
  CSubProblem(const CSubProblem&); // prohibited
public:
  CSubProblem();
  virtual ~CSubProblem();

  // set_femesh must be called right after initialization.  It's not
  // called by the constructor because the constructor is called by a
  // RegisteredClass registration, before the mesh is known.
  void set_femesh(FEMesh *msh) { mesh = msh; }
  FEMesh *mesh;

  Lock precomputeLock;
  bool precomputeRequired;

  CMicrostructure *get_microstructure() const;

  // redefined() must be called when the criteria for choosing nodes
  // and elements in a subproblem has changed.  For example, if
  // Material assignments have changed,
  // MaterialSubProblem::redefined() should be called.
  virtual void redefined() {}

  virtual ElementIterator element_iterator() = 0;
  virtual NodeIterator node_iterator() = 0;
  virtual FuncNodeIterator funcnode_iterator() = 0;
  virtual bool contains(const Element *) = 0;
  virtual bool containsNode(const Node *) const = 0;

  std::vector<int> freedofmap;
  std::vector<int> fixeddofmap;
  std::vector<int> indepeqnmap;
  std::vector<int> depeqnmap;

#ifdef HAVE_MPI
  // These are arrays corresponding to the freedofmap and indepeqnmap
  // but with the shared non-owned rows and columns indicated.  At the
  // end of the day, these should map the indices of the local
  // stiffness matrix onto the combined stiffness matrix
  std::vector<int> m_precombined_freedofmap;
  std::vector<int> m_precombined_indepeqnmap;
  // These are less than or equal to m_precombined_freedofmap.size()
  // (not counting the -1)
  int m_precombined_dofsize;
  int m_precombined_eqnsize;
  void set_parallel_mapping();
  // This member stores the cumulative sum of pdofsizelist
  // (dofsize0+dofsize1+...)
  int m_combinedmatrixdim;
  // This function for displaying solution at the FuncNodes
  void NodalPositionSolution(double* temp_array);
  //
  void ResetSharesData();
  // Gather total number of unique nodes for this subproblem.
  // First get the number of nodes in this process for this
  // subproblem (if the node is shared, count it if this
  // process owns the node). When these are collected
  // from all processes and added together, it will give the total number
  // of unique nodes for this subproblem (useful for subproblem info)
  int GatherNumNodes();
#endif

//   void dump_dof() const;	// for debugging
//   void dump_eqn() const;

  void rcmapper(bool (CSubProblem::*)(int) const,
		std::vector<int>&, int) const; 
  void mapdofeqs();  // Creates the matrix-partitioning maps.
  void resetdofeqs(); // Undoes BC invocation, sets dofs/eqns to free/indep.
  void cleanlists();
  void clear_brhs();

  CSRmatrix Amatrix() const;
  CSRmatrix Bmatrix() const;
  CSRmatrix Cmatrix() const;
  CSRmatrix Cpmatrix() const;

  LinearizedSystem linearsystem;
  void make_linear_system();
  const SparseLinkMatrix<double> &stiffness_matrix() const;
  const SparseLinkMatrix<double> &conductivity_matrix() const;
  const SparseLinkMatrix<double> &mass_matrix() const;
  VECTOR_D &boundary_rhs() const;

  std::vector<int> A_rowblocks() const;
  std::vector<int> A_colblocks() const;

  void get_unknowns(VECTOR_D&, const std::vector<int> &) const;
  void set_unknowns(const VECTOR_D&, const std::vector<int>&);

  void get_rhs(VECTOR_D&, const std::vector<int>&) const;
//   void set_rhs(const VECTOR_D&, const std::vector<int>&); // not used?

  void activate_equation(Equation &);
  void deactivate_equation(Equation &);
  bool is_active_equation(const Equation&) const;
  int n_active_eqns() const;

  void kinetic_activate_equation(Equation &);
  void kinetic_deactivate_equation(Equation &);
  bool is_kinetically_active_equation(Equation&);
  bool kinetics_activated();

  void activate_dynamics(Equation &);
  void deactivate_dynamics(Equation &);
  bool is_dynamically_active_equation(Equation&);
  bool dynamics_activated();

  void activate_flux(const Flux&);
  void deactivate_flux(const Flux&);
  bool is_active_flux(const Flux&) const;

  // activate, etc, for Fields have to be routed through Field virtual
  // functions, because they have to know if the Field is a
  // CompoundField or not.  The routines that do the real work are
  // SubProblem::do_activate, etc.
  void activate_field(Field&);
  void deactivate_field(Field&);
  void acquire_field_data(Field &field, const CSubProblem*);
  Field &define_field(Field&);
  void undefine_field(Field&);

//   void set_in_plane(const Field&, bool);
  bool in_plane(const Field&) const;

  // active fields are being solved for
  void do_activate_field(Field&);
  void do_deactivate_field(Field&);
  bool is_active_field(const Field&) const;
  int n_active_fields() const;

  // defined fields have values at nodes
  void do_define_field(Field&);
  void do_undefine_field(Field&);
  bool is_defined_field(const Field&) const;

  // List of this subproblem's defined fields.
  std::vector<CompoundField*>* all_compound_fields() const;
  std::vector<Equation*>* all_equations() const;
  std::vector<Flux*>* all_fluxes() const;
  std::vector<Flux*> allFluxes() const;  // no need to free memory.

  // Call requirePrecompute() when Materials change or Fields are
  // defined.  The actual precompute() function is in csubproblem.spy.
  void requirePrecompute();

  // Functions for symmetrizing the stiffness matrix, if
  // possible. set_equation_mapping is called from the conjugacy code,
  // and sets up a map whose index is the position of a nodal equation
  // in the FEMesh's master list of nodal equations, and whose value
  // is the row of the master stiffness matrix to which this
  // corresponds. "map_forward_index" should be used to get the row of
  // the stiffness matrix corresponding to a given equation,
  // component, and node.  "map_forward_equation" is similar, except
  // it returns a reference to the actual nodal equation object for
  // that row.  set_trivial_equation_mapping is used in the case where
  // we have inconsistent conjugacy data, but which to proceed
  // anyways.  
  void set_trivial_equation_mapping();
  void set_equation_mapping(const std::vector<CConjugatePair*>*);
  int map_forward_index(const Equation&, const FuncNode&, const int&) const;
  const NodalEquation & map_forward_equation(const Equation&, 
					     const FuncNode&, 
					     const int&) const;
  // Constructs and uses the forward_mapping object to transfer row
  // entries from the unsymmetric matrix to the symmetric one.  It
  // simultaneously builds the backward mapping, which will restore
  // the degrees of freedom values.  TODO: Is this an obsolete
  // comment, or is it just misplaced?
  void clear_symmetricity_maps();
  int check_symmetricity(const CSRmatrix &); //for debugging only
  // end of functions to symmetrize stiffness matrix

  // Run post-solver processing routines in those materials/properties
  // that have them.  This should eventually depend upon which
  // auxiliary fields are contributed to by the material.
  void post_process();

  // API for setting/referring to the read-write lock.  Set_rwlock
  // should be called exactly once when the SubProblem is inserted
  // into a SubProblemContext object.
  void set_rwlock(RWLock *rw) { rwlock = rw; };
  RWLock *get_rwlock() { return rwlock;};

private:

  // Is a Flux active?  Stored as an int, because a Flux can be
  // activated more than once.
  std::vector<int> active_flux;
  
  // Machinery to symmetrize the stiffness matrix
  std::vector<int> *global_equation_map;

  // This is data that can't live in the Field class, because there is
  // only one instance of each Field, but it might have different
  // properties on each SubProblem, and there might be more than one
  // SubProblem.  Planarity data isn't stored here, because it's
  // determined at the FEMesh level.
  class FieldData {
  public:
    FieldData();
  private:
    bool active;		// is this field being solved for?
    bool defined;		// have DoF's been assigned?
    friend class CSubProblem;
  };
  std::vector<FieldData> fielddata; 

  // This is data that can't live in the Equation class, for the same
  // reason that the FieldData can't live in the Field class.
  class EquationData {
  public:
    EquationData();
  private:
    bool active;
    bool kin_active;
    bool dyn_active;
    friend class CSubProblem;
  };
  std::vector<EquationData> eqndata;

  int n_active_eqn;		// number of explicitly activated eqns
  int n_active_field;		// number of explicitly activated fields

  // Routines used to pick rows and columns of the stiffness matrix
  // when extracting submatrices
  bool solve_for_this_dof(int) const;
  bool solve_this_nodal_eqn(int) const;
  bool dont_solve_for_this_dof(int) const;
  bool dont_solve_this_nodal_eqn(int) const;
  void cleanmap(std::vector<int>&);       // Clean up a particular map.
  // cleanmap is called by public method "cleanlists".

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
  //time dependent machinery starts here
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
private:
  double local_time;
public:
  void set_time(const double the_time) {local_time = the_time;}
  void update_time(const double dt) {local_time += dt;}
  double current_time() const {return local_time;}
  CSRmatrix ConductivityMatrix() const;
  CSRmatrix MassMatrix() const;
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
  //time dependent machinery ends here
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

  friend class SubProblemElementIterator;
  friend class SubProblemFuncNodeIterator;
  friend class SubProblemNodeIterator;

  //Adaptive Mesh Refinement stuff translocated from femesh.h
public:
  // TODO: encapsulate all of the ZZ error estimation stuff, so that
  // we can switch estimators.

  // create & add a new CSCPatch pointer
  void init_scpatches(const std::vector<int>*);
  void add_scpatch(const int, const Material*, 
		   const int,
		   const std::vector<int>*,
		   const std::vector<int>*,
		   const int);
  // getting elements & nodes from the patch
  std::vector<int> *get_elements_from_patch(const int, const Material*);
  std::vector<int> *get_nodes_from_patch(const int, const Material*);
  // recovering flux(es)
  void init_nodalfluxes();
  void recover_fluxes();
  // adding recovered flux
  void add_this_flux(const Material*, const Flux*,
		     const Node*, VECTOR_D*);
  // recovered flux at a given point
  VECTOR_D *get_recovered_flux(const Flux*, const Element*,
			       const MasterCoord&);
  // reporting recovered fluxes at a given point -- debug purpose
  void report_recovered_fluxes(const Element*, const Coord*);
  // estimating error
  double zz_L2_estimate(const Element*, const Flux*);
  void zz_L2_estimate_sub(const Element*, const Flux*,
			  const int&, double&, double&,
			  const MasterCoord&, const double&);
  std::vector<double> *zz_L2_weights(const Flux*,
				     const double&, const double&);
  void zz_L2_weights_sub(const Element*, const Flux*,
			  const int&, double&,
			  const MasterCoord&, const double&);

private:
  // Storage for CSCPatch's (keyed an assembly node)
  // NodalSCPatches contains as many CSCPatches as no. of Materials
  // at the assembly node.
  std::map<const int, NodalSCPatches*> scpatches;
  // Storage for SCpatch Recovered Fluxes
  std::map<const int, NodalFluxes*> recovered_fluxes;
};				// class CSubProblem

#endif // SUBPROBLEM_H
