# -*- python -*-
# $RCSfile: subproblemcontext.py,v $
# $Revision: 1.23 $
# $Author: vrc $
# $Date: 2007-09-26 22:02:56 $


# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from ooflib.SWIG.common import config
from ooflib.SWIG.common import switchboard
from ooflib.SWIG.common import timestamp
from ooflib.SWIG.engine import field
from ooflib.SWIG.engine import flux
from ooflib.common import debug
from ooflib.common.IO import parameter
from ooflib.common.IO import whoville
from ooflib.engine import mesh
from ooflib.engine import symstate
import string

class SubProblemContext(whoville.Who):
    def __init__(self, name, classname, obj, parent, secretFlag=0):
        whoville.Who.__init__(self, name, classname, obj, parent, secretFlag)
        obj.set_rwlock(self.rwLock)
        obj.set_mesh(parent)

        self.solverStatus = None
        self.solverMessage = None
        self.solver = None
        self.defnchanged = timestamp.TimeStamp() # fields or eqns changed
        self.defnchanged.backdate()
        self.solutiontimestamp = timestamp.TimeStamp() # last "solve" command
        self.solutiontimestamp.backdate()

        self.requestCallback(("preremove who", "SubProblem"),
                             self.preremoveCB)
        self.requestCallback("subproblem redefined", self.redefinedCB)

        self.mesh_symmetry = symstate.INCONSISTENT

        # Subproblems are both Who objects and RegisteredClass
        # objects.  If their Registration has a callable 'startup'
        # member, it's invoked here.  This can be used to set up
        # switchboard signals, for instance.  A corresponding
        # 'cleanup' function is called when the subproblem is
        # destroyed.
        reg = obj.getRegistration()
        if hasattr(reg, 'startup'):
            reg.startup(self)
            
    def preremoveCB(self, path):
        # If one of our dependencies is being removed, we're done for.
        # We might as well end it all right now.
        path = string.join(path, ':')
        if path in self.getObject().get_dependencies():
            self.reserve()
            self.begin_writing()
            try:
                self.destroy()
            finally:
                self.end_writing()
                self.cancel_reservation()

    def destroy(self):
        # See 'startup' comment in __init__.
        reg = self.getObject().getRegistration()
        if hasattr(reg, 'cleanup'):
            reg.cleanup(self)
        # Undefine all of our fields and deactivate equations, so that
        # they'll be removed from the Mesh if they're no longer
        # needed.
        for field in self.all_compound_fields():
            self.getObject().undefine_field(field)
        for eqn in self.all_equations():
            self.getObject().deactivate_equation(eqn)
        subproblems.remove(self.path())

    def getMicrostructure(self):
        return self.getParent().getMicrostructure()

    def nelements(self):
        return self.getObject().element_iterator().size()

    def nnodes(self):
        return self.getObject().node_iterator().size()

    def nfuncnodes(self):
        return self.getObject().funcnode_iterator().size()

    def area(self):
        a = 0.0
        iter = self.getObject().element_iterator()
        while not iter.end():
            a += iter.element().area()
            iter.next()
        return a

    def solved(self):
        switchboard.notify("subproblem changed", self)

    def setSolverStatus(self, success, solver, message):
        self.solverStatus = success or 'failed'
        self.solver = solver
        self.solverMessage = message or ''

    def changed(self):
        # Called when fields, equations, or materials have changed.
        self.defnchanged.increment()    # timestamp
        self.getObject().requirePrecompute()
        switchboard.notify("subproblem changed", self)

    def redefined(self):
        # Called when the things affecting the subproblem definition
        # itself have changed.  For example, a MaterialSubProblem is
        # redefined when its Material is assigned to or removed from
        # pixels.
        self.getObject().redefined()
        switchboard.notify("subproblem redefined", self.path())

    def redefinedCB(self, subppath):
        for dep in self.getObject().get_dependencies():
            if dep == subppath:
                self.redefined()

    def has_solution(self):
        return self.solutiontimestamp > max(self.defnchanged,
                                            self.parent.getOwnTimeStamp())

    def getTimeStamp(self, gfxwindow):
        ## This does *not* check the timestamp of the parent
        ## meshcontext, because the mesh's getTimeStamp function
        ## checks the timestamps of all of its subproblems.  That's
        ## because changes in a subproblem must cause a mesh to be
        ## redrawn.
        return max(self.defnchanged, self.solutiontimestamp)

    def all_compound_fields(self):
        # Because Fields aren't PythonExportable objects,
        # FEMesh.all_compound_fields() returns a list of CompoundField
        # objects without the correct subclass associations.  However,
        # field names must be unique, so we can get the correct field
        # with the full class hierarchy by looking it up in the list
        # maintained by the field module.
        fields = self.getObject().all_compound_fields()
        return [field.allFields[f.name()] for f in fields]

    def all_fields(self):
        fields = self.getObject().all_fields()
        return [field.allFields[f.name()] for f in fields]
    
    def all_equations(self):
        return self.getObject().all_equations()
    def all_equations_bc(self):
        eqns =  [eq for eq in self.getObject().all_equations()
                if eq.allow_boundary_conditions()]
        return eqns

    def all_fluxes(self):
        fluxes = self.getObject().all_fluxes()
        return [flux.allFluxes[f.name()] for f in fluxes]

    def is_defined_field(self, field):
        return field.is_defined(self.getObject())

    # This function must be called after fields or equations are added
    # to or removed from this subproblem -- BCs which have become
    # incomputable will shut themselves off, and those which have
    # become computable will switch themselves on.
    def autoenableBCs(self):
        self.reserve()
        self.begin_writing()
        try:
            for (name, bc) in self.parent.allBoundaryConds():
                bc.auto_enable(self)
        finally:
            self.end_writing()
            self.cancel_reservation()
        # Mild abuse of this switchboard callback, with placeholder args.
        switchboard.notify("boundary conditions changed", self, None, True)


    # Function to assess the well-posedness of the current state.
    # Returns None if the problem is well-posed, and returns a message
    # string indicating the problem if it's not well-posed.
    def check_materials(self):
        status = None
        badmatls = self.getObject().check_materials()
        badeqns = self.getObject().check_equations()
        if badmatls or badeqns:
            status= "Ill-posed problem!\n"
            if badmatls:
                status += "Badly defined material%s!\n" % \
                          ("s"*(len(badmatls)>1))
                for matl in badmatls:
                    status += string.join(
                        matl.consistency_messages(), '\n')
            if badeqns:
                status += "Equation%s with no flux contributions!" % \
                          ("s"*(len(badeqns)>1))
                for eqn in badeqns:
                    status +="\n    " + eqn.name()
        return status

    def materialsConsistent(self):
        # check_materials returns a list of inconsistent materials
        return (not self.getObject().check_materials()) and \
               (not self.getObject().check_equations())

    def status(self):
        if self.solutiontimestamp > max(self.parent.getOwnTimeStamp(),
                                        self.defnchanged):
            status = "Stiffness Matrix: " + self.mesh_symmetry + "\n" + \
                     "Solver: " + `self.solver` + '\n' + \
                     "Solution Status: " + self.solverStatus + "\n" + \
                     "---------------\n" + \
                     "Message:\n" + self.solverMessage
        else:
            material_error = self.check_materials()
            if material_error:
                status = material_error
            else:
                status = "Stiffness Matrix: " + self.mesh_symmetry + "\n" + \
                         'Unsolved' 
        if config.devel():
            status += "\ntime index = " + str(self.getObject().current_time())
        return status

    def make_linear_system(self, force=False):
        if ((self.meshchanged > self.solutiontimestamp) or 
            (self.parent.materialschanged > self.solutiontimestamp) or 
            self.getObject().nonlinear_materials() or force):
            import time
            start = time.clock()
            self.getObject().make_linear_system()
            debug.fmsg("Linear system constructed in", time.clock()-start, "sec")
    def get_symmetry_state(self):
        return self.mesh_symmetry
    def set_symmetry_state(self, state):
        ## state should be one of the strings defined in symstate.py.
        self.mesh_symmetry = state


#####################################################

subproblems = whoville.WhoClass(
    'SubProblem',
    instanceClass=SubProblemContext,
    ordering=400,
    secret=True,
    parentClass=mesh.meshes)

#####################################################

# SubProblemParameters are used instead of WhoParameters in cases in
# which only the subproblem name, and not its full path, needs to be
# specified.  The associated gui widget lists only the subproblems
# belonging to the mesh that's in the same scope.

# TODO: It would be good to use a variant of the WhoParameter for
# this.  The difficulty with that is that its checker() method would
# have to find the correct Mesh to know which SubProblem names were
# valid.  But for creating compound subproblems, the Mesh is given not
# in the SubProblem constructor arguments, but one level up:
##  OOF.Subproblem.New(mesh, UnionSubProblem('subOne', 'subTwo'))

class SubProblemParameter(parameter.StringParameter):
    pass
